@@ -27,6 +27,7 @@ import apt_pkg
 import ConfigParser
 import copy
 import datetime
+import fcntl
 import re
 import os
 import string
@@ -93,6 +94,7 @@ class UnattendedUpgradesCache(apt.Cache):
                 logging.debug("adjusting candidate version: '%s'" % new_cand)
                 pkg.candidate = new_cand
 
+
 class LogInstallProgress(apt.progress.base.InstallProgress):
     """ Install progress that writes to self.LOG
         (/var/run/unattended-upgrades.progress by default)
@@ -104,7 +106,46 @@ class LogInstallProgress(apt.progress.base.InstallProgress):
         f=open(self.LOG, "w")
         f.write(_("Progress: %s %% (%s)") % (percent, pkg))
         f.close()
-        
+
+    def _fixup_fds(self):
+        required_fds = [ 0, 1, 2, # stdin, stdout, stderr
+                         self.writefd,
+                         self.write_stream.fileno(),
+                         self.statusfd,
+                         self.status_stream.fileno()
+                       ]
+        # ensure that our required fds close on exec
+        for fd in required_fds[3:]:
+            old_flags = fcntl.fcntl(fd, fcntl.F_GETFD)
+            fcntl.fcntl(fd, fcntl.F_SETFD, old_flags | fcntl.FD_CLOEXEC)
+        # close all fds
+        proc_fd = "/proc/self/fd"
+        if os.path.exists(proc_fd):
+            error_count = 0
+            for fdname in os.listdir(proc_fd):
+                try:
+                    fd = int(fdname)
+                except Exception as e:
+                    print "ERROR: can not get fd for '%s'" % fdname
+                if fd in required_fds:
+                    continue
+                try:
+                    os.close(fd)
+                    #print "closed: ", fd
+                except OSError as e:
+                    # there will be one fd that can not be closed
+                    # as its the fd from pythons internal diropen()
+                    # so its ok to ignore one close error
+                    error_count += 1
+                    if error_count > 1:
+                        print "ERROR: os.close(%s): %s" % (fd, e)
+
+    def fork(self):
+        pid = os.fork()
+        if pid == 0:
+            self._fixup_fds()
+        return pid
+
 
 class Unlocked:
     """ context manager for unlocking the apt lock while cache.commit()
@@ -163,6 +204,9 @@ def get_allowed_origins_legacy():
             (distro_id, distro_codename) = s.split(':')
         else:
             (distro_id, distro_codename) = s.split()
+        # escape "," (see LP: #824856) - i wonder if there is a simpler way?
+        distro_id = re.sub(r'([^\\]),', r'\1\\,', distro_id)
+        distro_codename = re.sub(r'([^\\]),', r'\1\\,', distro_codename)
         # convert to new format
         allowed_origins.append("o=%s,a=%s" % (substitute(distro_id), 
                                               substitute(distro_codename)))
@@ -210,25 +254,48 @@ def match_whitelist_string(whitelist, origin):
             res &= (value == origin.site)
     return res
 
-def upgrade_normal(cache, pkgs_to_upgrade, logfile_dpkg):
+
+def cache_commit(cache, logfile_dpkg, verbose, iprogress=None):
+    """Commit the changes from the given cache to the system"""
+    # set debconf to NON_INTERACTIVE, redirect output
+    os.putenv("DEBIAN_FRONTEND", "noninteractive")
+
+    # only act if there is anything to do (important to avoid writing
+    # empty log stanzas)
+    if len(cache.get_changes()) == 0:
+        return True, None
+
     error = None
     res = False
-    iprogress = LogInstallProgress()
+    if iprogress is None:
+        iprogress = LogInstallProgress()
+
     try:
         with Unlocked():
             res = cache.commit(install_progress=iprogress)
-    except SystemError,e:
+    except SystemError as e:
         error = e
+        if verbose:
+            logging.exception("Exception happened during upgrade.")
+    return res, error
+
+
+def upgrade_normal(cache, logfile_dpkg, verbose=False):
+    res, error = cache_commit(cache, logfile_dpkg, verbose)
     if res:
         logging.info(_("All upgrades installed"))
     else:
         logging.error(_("Installing the upgrades failed!"))
-        logging.error(_("error message: '%s'") % error)
-        logging.error(_("dpkg returned a error! See '%s' for details") % \
-                          logfile_dpkg)
+        logging.error(_("error message: '%s'"), error)
+        logging.error(_("dpkg returned a error! See '%s' for details"), \
+                      logfile_dpkg)
     return res
 
 def upgrade_in_minimal_steps(cache, pkgs_to_upgrade, logfile_dpkg=""):
+
+    install_log = LogInstallProgress()
+    install_log.LOG += ".minimal-steps"
+
     # setup signal handler
     signal.signal(signal.SIGUSR1, signal_handler)
 
@@ -259,26 +326,22 @@ def upgrade_in_minimal_steps(cache, pkgs_to_upgrade, logfile_dpkg=""):
             cache.clear()
 
         # write progress log information
-        f=open(PROGRESS_LOG+".minimal-steps", "w")
-        percent = len(to_upgrade)/len(pkgs_to_upgrade)*100.0
-        f.write(_("Progress: %s %%: (%s)") % (
-                    percent, ",".join(smallest_partition)))
-        f.close()
-
+        if len(pkgs_to_upgrade) > 0:
+            percent = (len(pkgs_to_upgrade)-len(to_upgrade)) / float(len(pkgs_to_upgrade))*100.0
+        else:
+            percent = 100.0
+        install_log.status_change(pkg=",".join(smallest_partition),
+                                  percent=percent,
+                                  status="")
         # apply changes
         logging.debug("applying set %s" % smallest_partition)
         rewind_cache(cache, [cache[name] for name in smallest_partition])
 
-        try:
-            iprogress = LogInstallProgress()
-            with Unlocked():
-                res = cache.commit(install_progress=iprogress)
-                if not res:
-                    raise Exception("cache.commit() returned false")
-            cache.open()
-        except Exception, e:
+        verbose = False
+        res, error = cache_commit(cache, logfile_dpkg, verbose, install_log)
+        if error:
             logging.error(_("Installing the upgrades failed!"))
-            logging.error(_("error message: '%s'") % e)
+            logging.error(_("error message: '%s'") % error)
             logging.error(_("dpkg returned a error! See '%s' for details") % \
                               logfile_dpkg)
             return False
@@ -306,6 +369,11 @@ def check_changes_for_sanity(cache, allowed_origins, blacklist):
             logging.debug("pkg '%s' now marked delete" % pkg.name)
             return False
         if pkg.marked_install or pkg.marked_upgrade:
+            # apt will never fallback from a trusted to a untrusted
+            # origin so its good enough if we have a single trusted one
+            if not any([o.trusted for o in pkg.candidate.origins]):
+                logging.debug("pkg '%s' is untrusted" % pkg.name)
+                return False
             if not is_allowed_origin(pkg.candidate, allowed_origins):
                 logging.debug("pkg '%s' not in allowed origin" % pkg.name)
                 return False
@@ -362,6 +430,10 @@ def conffile_prompt(destFile, prefix=""):
                     # ignore if conffile is obsolete or does not exist
                     if obs == "obsolete" or not os.path.exists(prefix+conf_file):
                         continue
+                    # ignore state "newconffile" until its clearer if there
+                    # might be a dpkg prompt (LP: #936870)
+                    if md5 == "newconffile":
+                        continue
                     # get conffile value from pkg, its ok if the new version
                     # does not have conffiles anymore
                     deb = apt_inst.DebFile(destFile)
@@ -446,7 +518,7 @@ def setup_apt_listchanges(conf="/etc/apt/listchanges.conf"):
     os.environ["APT_LISTCHANGES_FRONTEND"] = "none"
 
 def send_summary_mail(pkgs, res, pkgs_kept_back, mem_log, logfile_dpkg):
-    " send mail (if configured in Unattended-Upgrades::Mail) "
+    " send mail (if configured in Unattended-Upgrade::Mail) "
     email = apt_pkg.config.find("Unattended-Upgrade::Mail", "")
     if not email:
         return
@@ -478,16 +550,64 @@ def send_summary_mail(pkgs, res, pkgs_kept_back, mem_log, logfile_dpkg):
         s += " " + wrap(" ".join(pkgs_kept_back), 70, " ")
         s += "\n"
     s += "\n"
-    s += _("Package installation log:")+"\n"
-    s += open(logfile_dpkg).read()
-    s += "\n\n"
+    if os.path.exists(logfile_dpkg):
+        s += _("Package installation log:")+"\n"
+        s += open(logfile_dpkg).read()
+        s += "\n\n"
     s += _("Unattended-upgrades log:\n")
     s += mem_log.getvalue()
     mail.stdin.write(s)
     mail.stdin.close()
     ret = mail.wait()
     logging.debug("mail returned: %s" % ret)
+
+
+def do_install(cache, pkgs_to_upgrade, options, logfile_dpkg):
+    # set debconf to NON_INTERACTIVE, redirect output
+    os.putenv("DEBIAN_FRONTEND","noninteractive");
+    setup_apt_listchanges()
     
+    # redirect to log
+    REDIRECT_INPUT = os.devnull
+    fd = os.open(REDIRECT_INPUT, os.O_RDWR)
+    os.dup2(fd,0)
+
+    logging.info(_("Writing dpkg log to '%s'") % logfile_dpkg)
+
+    # do not create log in dry-run mode, just output to stdout/stderr
+    if options.dry_run:
+        old_stdout = 1
+        old_stderr = 2
+    else:
+        fd = os.open(logfile_dpkg, os.O_RDWR|os.O_CREAT, 0644)
+        old_stdout = os.dup(1)
+        old_stderr = os.dup(2)
+        os.dup2(fd,1)
+        os.dup2(fd,2)
+
+    try:
+        if (options.minimal_upgrade_steps or 
+            # COMPAT with the mispelling
+            apt_pkg.config.find_b("Unattended-Upgrades::MinimalSteps", False) or
+            apt_pkg.config.find_b("Unattended-Upgrade::MinimalSteps", False)):
+            open("/var/run/unattended-upgrades.pid", "w").write("%s" % os.getpid())
+            # try upgrade all "pkgs" in minimal steps
+            pkg_install_success = upgrade_in_minimal_steps(
+                cache, [pkg.name for pkg in pkgs_to_upgrade], logfile_dpkg)
+        else:
+            pkg_install_success = upgrade_normal(
+                cache, logfile_dpkg, options.debug)
+    except Exception as e:
+        # print unhandled exceptions here this way, while stderr is redirected
+        os.write(old_stderr, "Exception: %s" % e)
+        pkg_install_success = False
+
+    # restore
+    os.dup2(old_stdout, 1)
+    os.dup2(old_stderr, 2)
+
+    return pkg_install_success
+
 
 def _setup_alternative_rootdir(rootdir):
     # clear system unattended-upgrade stuff
@@ -500,6 +620,10 @@ def _setup_alternative_rootdir(rootdir):
     if os.path.isdir(rootdir+"/etc/apt/apt.conf.d"):
         apt_pkg.read_config_dir(apt_pkg.config,
                                 rootdir + "/etc/apt/apt.conf.d")
+    logdir = os.path.join(rootdir, "var", "log", "unattended-upgrades")
+    if not os.path.exists(logdir):
+        os.makedirs(logdir)
+    apt.apt_pkg.config.set("Unattended-Upgrade::LogDir", logdir)
 
 def _get_logdir():
     logdir= apt_pkg.config.find_dir(
@@ -528,6 +652,32 @@ def _setup_logging(options):
                         filename=logfile)
 
 
+def get_auto_removable(cache):
+    return set([pkg.name for pkg in cache
+                if pkg.is_auto_removable])
+
+
+def do_auto_remove(cache, options, logfile_dpkg, verbose=False):
+    auto_removable = get_auto_removable(cache)
+
+    for pkgname in auto_removable:
+        logging.debug("marking %s for remove" % pkgname)
+        cache[pkgname].mark_delete()
+    logging.info(_("Packages that are auto removed: '%s'")
+                 % " ".join(auto_removable))
+
+    # do it
+    res, error = cache_commit(cache, logfile_dpkg, verbose)
+    if res:
+        logging.info(_("Packages auto-removed"))
+    else:
+        logging.error(_("Auto-removing the packages failed!"))
+        logging.error(_("Error message: '%s'") % error)
+        logging.error(_("dpkg returned an error! See '%s' for details")
+                      % logfile_dpkg)
+    return res
+
+
 def main(options, rootdir=""):
 
     # useful for testing
@@ -603,8 +753,6 @@ def main(options, rootdir=""):
     # find out about the packages that are upgradable (in a allowed_origin)
     pkgs_to_upgrade = []
     pkgs_kept_back = []
-    pkgs_auto_removable = set([pkg.name for pkg in cache 
-                               if pkg.is_auto_removable])
 
     # now do the actual upgrade
     for pkg in cache:
@@ -616,7 +764,16 @@ def main(options, rootdir=""):
                 pkg.mark_upgrade()
                 if check_changes_for_sanity(cache, allowed_origins,
                                             blacklisted_pkgs):
+                    # add to packages to upgrade
                     pkgs_to_upgrade.append(pkg)
+                    # re-eval pkgs_kept_back as the resolver may fail to 
+                    # directly upgrade a pkg, but that may work during
+                    # a subsequent operation, see debian bug #639840
+                    for pkgname in pkgs_kept_back:
+                        if (cache[pkgname].marked_install or
+                            cache[pkgname].marked_upgrade):
+                            pkgs_kept_back.remove(pkgname)
+                            pkgs_to_upgrade.append(cache[pkgname])
                 else:
                     logging.debug("sanity check failed")
                     rewind_cache(cache, pkgs_to_upgrade)
@@ -628,6 +785,7 @@ def main(options, rootdir=""):
                 pkgs_kept_back.append(pkg.name)
                 
 
+    pkgs_to_upgrade.sort(key=lambda p: p.name)
     pkgs = "\n".join([pkg.name for pkg in pkgs_to_upgrade])
     logging.debug("pkgs that look like they should be upgraded: %s" % pkgs)
            
@@ -702,21 +860,29 @@ def main(options, rootdir=""):
     else:
         logging.debug("dpkg is configured not to cause conffile prompts")
 
-    # do auto-remove
-    if apt_pkg.config.find_b("Unattended-Upgrade::Remove-Unused-Dependencies", False):
-        now_auto_removable = set([pkg.name for pkg in cache 
-                               if pkg.is_auto_removable])
-        for pkgname in now_auto_removable-pkgs_auto_removable:
-            logging.debug("marking %s for remove" % pkgname)
-            cache[pkgname].mark_delete()
-        logging.info(_("Packages that are auto removed: '%s'") %
-                     " ".join(now_auto_removable-pkgs_auto_removable))
+    # auto-removal
+    if apt_pkg.config.find_b(
+            "Unattended-Upgrade::Remove-Unused-Dependencies", False):
+        pending_autoremovals = get_auto_removable(cache)
+    else:
+        pending_autoremovals = []
 
     logging.debug("InstCount=%i DelCount=%i BrokenCout=%i" % (cache._depcache.inst_count, cache._depcache.del_count, cache._depcache.broken_count))
 
     # exit if there is nothing to do and nothing to report
-    if (len(pkgs_to_upgrade) == 0) and (len(pkgs_kept_back) == 0):
-        logging.info(_("No packages found that can be upgraded unattended"))
+    if (len(pending_autoremovals) == 0 and
+            len(pkgs_to_upgrade) == 0 and
+            len(pkgs_kept_back) == 0):
+        logging.info(_("No packages found that can be upgraded unattended "
+                       "and no pending auto-removals"))
+        return
+
+    # check if its configured for install on shutdown, if so, the 
+    # environment UNATTENDED_UPGRADES_FORCE_INSTALL_ON_SHUTDOWN will
+    # be set by the unatteded-upgrades-shutdown script
+    if (not "UNATTENDED_UPGRADES_FORCE_INSTALL_ON_SHUTDOWN" in os.environ and
+        apt_pkg.config.find_b("Unattended-Upgrade::InstallOnShutdown", False)):
+        logger.debug("Configured to install on shutdown, so exiting now")
         return
 
     # check if we are in dry-run mode
@@ -728,46 +894,41 @@ def main(options, rootdir=""):
     pkgs = " ".join([pkg.name for pkg in pkgs_to_upgrade])
     logging.info(_("Packages that are upgraded: %s" % pkgs))
 
-    # set debconf to NON_INTERACTIVE, redirect output
-    os.putenv("DEBIAN_FRONTEND","noninteractive");
-    setup_apt_listchanges()
-    
-    # redirect to log
-    REDIRECT_INPUT = os.devnull
-    fd = os.open(REDIRECT_INPUT, os.O_RDWR)
-    os.dup2(fd,0)
-
+    # get log
     now = datetime.datetime.now()
     logfile_dpkg = os.path.join(
         _get_logdir(), 'unattended-upgrades-dpkg_%s.log' % now.isoformat('_'))
-    logging.info(_("Writing dpkg log to '%s'") % logfile_dpkg)
-    fd = os.open(logfile_dpkg, os.O_RDWR|os.O_CREAT, 0644)
-    old_stdout = os.dup(1)
-    old_stderr = os.dup(2)
-    os.dup2(fd,1)
-    os.dup2(fd,2)
 
-    try:
+    # only perform install step if we actually have packages to install
+    pkg_install_success = None
+    shutdown_lock = -1
+    if len(pkgs_to_upgrade) > 0:
         # lock for the shutdown check - its fine if the system
         # is shutdown while downloading but not so much while installing
-        apt_pkg.get_lock("/var/run/unattended-upgrades.lock")
+        shutdown_lock = apt_pkg.get_lock("/var/run/unattended-upgrades.lock")
+        # do install
+        pkg_install_success = do_install(cache, pkgs_to_upgrade, options, logfile_dpkg)
 
-        if (options.minimal_upgrade_steps or 
-            apt_pkg.config.find_b("Unattended-Upgrades::MinimalSteps", False)):
-            open("/var/run/unattended-upgrades.pid", "w").write("%s" % os.getpid())
-            # try upgrade all "pkgs" in minimal steps
-            pkg_install_success = upgrade_in_minimal_steps(
-                cache, [pkg.name for pkg in pkgs_to_upgrade], logfile_dpkg)
-        else:
-            pkg_install_success = upgrade_normal(
-                cache, [pkg.name for pkg in pkgs_to_upgrade], logfile_dpkg)
-    except Exception, e:
-        # print unhandled exceptions here this way, while stderr is redirected
-        os.write(old_stderr, "Exception: %s" % e)
+    # remove unneeded packages if there are any
+    if apt_pkg.config.find_b(
+            "Unattended-Upgrade::Remove-Unused-Dependencies", False):
+        cache = UnattendedUpgradesCache(rootdir=rootdir,
+                                        allowed_origins=allowed_origins)
+        if cache._depcache.broken_count > 0:
+            print(_("Cache has broken packages, exiting"))
+            logging.error(_("Cache has broken packages, exiting"))
+            sys.exit(1)
 
-    # restore
-    os.dup2(old_stdout, 1)
-    os.dup2(old_stderr, 2)
+        remove_success = do_auto_remove(
+            cache, options, logfile_dpkg,
+            options.debug)
+        if pkg_install_success is not False:
+            pkg_install_success = remove_success
+
+    logging.debug("InstCount=%i DelCount=%i BrokenCount=%i"
+                  % (cache._depcache.inst_count,
+                     cache._depcache.del_count,
+                     cache._depcache.broken_count))
 
     # send a mail (if needed)
     if not options.dry_run:
@@ -777,6 +938,8 @@ def main(options, rootdir=""):
     # auto-reboot (if required and the config for this is set
     if (apt_pkg.config.find_b("Unattended-Upgrade::Automatic-Reboot", False) and
         os.path.exists(REBOOT_REQUIRED_FILE)):
+        if shutdown_lock > 0:
+            os.close(shutdown_lock)
         logging.warning("Found %s, rebooting" % REBOOT_REQUIRED_FILE)
         subprocess.call(["/sbin/reboot"])
         
